/*
  file:         Memory_routines.cc
  purpose:      routines to deal with allocation and freeing memory space.
                When Memory can not be allocated, No_memory_panic() is
                called, which issues No memory error and exits to prompt.
  created:      Sekhar Muddana circa 1990
  modified:     pasha jan 1 2019
  modification: minor code formatting
*/

#include <stdlib.h>
#include <iostream>
#include "Id_routines.hh"
#include "Po_parse_poly.hh"
#include "Po_prod_bst.hh"
#include "Po_parse_exptext.hh"

#include "Memory_routines.hh"

#define DB_MEM_ALLOC  0

#ifdef DMALLOC
#include <dmalloc.h>
#endif

void No_memory_panic();

static struct unexp_tnode* Free_tnode_queue = NULL;
static int Num_free_tnodes = 0;


// If there are some free tnodes on the queue, return pointer to first tnode on
// the queue and delete this tnode from the queue, else allocate space by 
// calling Mymalloc() and return a pointer to the allocated tnode.
// REQUIRES: Free_tnode_queue_ptr -- queue of free tnodes.           
struct unexp_tnode* Unexp_tnode_alloc()
{
  struct unexp_tnode *new_tnode;

#if DB_MEM_ALLOC
  static int i = 1;
  printf("in unexp alloc %d\n",i++);
#endif
    
  if (Free_tnode_queue != NULL) 
	{
	  new_tnode = Free_tnode_queue;
	  Free_tnode_queue = Free_tnode_queue->next;
	  Num_free_tnodes--;
    }
  else 
	{
	  new_tnode = ((struct unexp_tnode *) Mymalloc(sizeof(struct unexp_tnode)));
    }
  
  new_tnode->_operator = -1;
  new_tnode->s_letter = ' ';
  new_tnode->scalar_num = 0;
  new_tnode->operand1 = NULL; 
  new_tnode->operand2 = NULL; 
  new_tnode->operand3 = NULL; 
  new_tnode->next = NULL; 
  return (new_tnode);
}


void Free_tnode (struct unexp_tnode* Pntr)
{
#if DB_MEM_ALLOC
  static int i = 1;
  printf("in free tnode %d\n",i++);
#endif

  if (Num_free_tnodes < 2000) 
	{
	  Pntr->next = Free_tnode_queue;
	  Free_tnode_queue = Pntr;
	  Num_free_tnodes++;
	}
  else
	{
	  free (Pntr);
	}
}



// RETURNS: Pointer to a new PROD_TNODE
PROD_TREEPTR prod_talloc()
{
  PROD_TREEPTR temp = (PROD_TREEPTR)Mymalloc (sizeof(PROD_TREENODE));
  return (temp);
}


// RETURNS: Pointer to a new polynomial
struct polynomial* Poly_alloc()
{
  struct polynomial* poly = 
	((struct polynomial *) Mymalloc(sizeof(struct polynomial)));
  return (poly);
}


// RETURNS: Pointer to a new Term_node
struct term_node* Term_node_alloc()
{
  struct term_node* temp = 
	((struct term_node *) Mymalloc (sizeof(struct term_node)));
  temp->left = temp->right = NULL;
  temp->letter = ' ';
  temp->number = 0; 
  return (temp);
}


// RETURNS: Pointer to a new Term_head
struct term_head* Term_head_alloc()
{
  struct term_head* temp = 
	((struct term_head *) Mymalloc (sizeof(struct term_head)));
  temp->next = NULL;
  return (temp);
}


// args: size -- amount of space needs to be allocated
// RETURNS: Pointer to allocated block of size size
void* Mymalloc (size_t size)
{
  if (size <= 0) 
	{
	  std::cerr << std::flush << std::endl 
				<< " Severe. Trying to allocate 0 bytes. " << std::endl;
	  return (NULL);
    }
  void* temp = malloc (size);
  if (temp == NULL)
	{
	  No_memory_panic();
	}
  return (temp);
}



// RETURNS: Pointer to a new id_queue_node
struct id_queue_node* Id_queue_node_alloc()
{
  struct id_queue_node* id = 
	(struct id_queue_node *) Mymalloc (sizeof(struct id_queue_node));
  return (id);
}


// Issue a warning and exit to $ prompt
void No_memory_panic()
{
  std::cerr << std::endl << "Memory overflow." << std::endl;
}


// end of Memory_routines.cc
