/*
  file:         PerformSub.cc
  created:      David P Jacobs circa 1990
  modified:     Pasha Zusmanovich feb 18,20-21,24,27,mar 9,14,may 18 2016
  modification: - c++ file ext.
                - cleanup
				- templated Scalar
				- support for dmalloc
				- restored debug functions
*/

// Given an identity and a substitution record, we perform the actual 
// substitution in this module. We have to permute a variable in all possible 
// ways, i.e., if the degree of a variable is 4 then we have 24 permutations,
// resulting in 24 equations. The permutations are obtatined using recursion and
// selection sort. The equations are appended to the list of equations L. An 
// equation is obtained by substituting basis elements in place of variables in
// the equation and multiplying out all the basis elements whose products are 
// stored in the multiplication table. So new basis elements and new products 
// are determined by solving these equations.


#include <stdlib.h>
#include "albert.hh"

#include "PerformSub.hh"

#ifdef DMALLOC
#include <dmalloc.h>
#endif

int  GetOtherIndexToSwap (int);
void PrintPermutation (int, Perm);
void SortPermutation (int);

Perm Permutation;
int Permutation_length;


int Albert::FreePermutationList (Perm* Pl)
{
  assert_not_null(Pl);
  for (int i=0; i < this->Num_vars; i++)
	{
	  free (Pl[i]);
	}
  return (1);
}



Perm GetFirstPermutation (const int Permutation_length)
{
  Perm temp_perm = (int*)(Mymalloc (Permutation_length * sizeof(int)));
  if (temp_perm == NULL)
	{
	  return (NULL);
	}
  for (int i=0; i<Permutation_length; i++)
	{
	  temp_perm[i] = i+1;
	}
  return (temp_perm);
}



static int GetIndex()
{
  int i;
  bool index_changed = 1;
  int index = 0;
  while ((index < Permutation_length) && index_changed)
	{
	  index_changed = 0;
	  for (i=index; i < Permutation_length - 1; i++) 
		{
		  if (Permutation[i] < Permutation[i+1]) 
			{
			  index++;
			  index_changed = 1;
			  break;
            }
        }
    }
  return (index);
}



Perm GetNextPermutation (const Perm P, const int Pl)
{
  Permutation = P;
  Permutation_length = Pl;
  const int j = GetIndex();
  if (j == 0)
	{
	  return (NULL);
	}
  const int index1 = j - 1;
  const int index2 = GetOtherIndexToSwap (index1);
  int temp = Permutation[index1];
  Permutation[index1] = Permutation[index2];
  Permutation[index2] = temp;
  SortPermutation (index1 + 1);
  return (Permutation);
}




int GetOtherIndexToSwap (int index1)
{
  for (int i = Permutation_length-1; i>index1; i--)
	{
	  if (Permutation[i] > Permutation[index1])
		{
		  return (i);
		}
	}
  std::cout << "ERROR from GetOtherIndexToSwap()" << std::endl;
  return (-1);
} 


// Selection Sort
void SortPermutation (int begin_index)
{
  int j, min_index, temp;
  for (int i = begin_index; i < Permutation_length - 1; i++) 
	{
	  min_index = i;
	  for (j=i+1; j < Permutation_length; j++)
		{
		  if (Permutation[j] < Permutation[min_index])
			{
			  min_index = j;
			}
		}
	  if (min_index != i) 
		{
		  temp = Permutation[i];
		  Permutation[i] = Permutation[min_index];
		  Permutation[min_index] = temp;
		}
	}
}



#if DEBUG_PERMUTATIONS
void Albert::PrintPermutationList() const
{
  assert_not_null_retvoid (this->Permutation_list);
  std::cout << "Permutation List is :" << std::endl;
  for (int i=0; i < this->Num_vars; i++) 
	{
	  this->PrintPermutation (i, this->Permutation_list[i]);
	  std::cout << std::endl;
	}
}

void Albert::PrintPermutation (const int Var_num, const Perm P) const
{
  for (int i=0; i < this->Deg_var[Var_num]; i++)
	{
	  std::cout << P[i];
	}
}
#endif

// end of PerformSubs.cc
