/*
  file:         Po_create_poly.c
  purpose:      routines needed to create a polynomial,
                from the input string entered by user
  created:      Sekhar Muddana circa 1990
  modified:     pasha jan 4 2019
  modification: Str_cat() relocated here
*/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <iostream>
#include "Build_defs.hh"
#include "Memory_routines.hh"
#include "Po_expand_poly.hh"
#include "Po_parse_poly.hh"	
#include "Po_semantics.hh"	
#include "Po_prod_bst.hh"	
#include "Po_parse_exptext.hh"	

#include "Po_create_poly.hh"	

#ifdef DMALLOC
#include <dmalloc.h>
#endif

void Create_exp_str      (const unexp_tnode*, char**, size_t*);
bool Found_minus         (const unexp_tnode*);
void Free_tnode_tree     (unexp_tnode*);
static void Str_cat (char**, const char[], size_t*);

#if DEBUG_EXP
void Print_tree          (const unexp_tnode*);
void Print_art_word      (const unexp_tnode*);
void Print_opr_prod_tree (const unexp_tnode*);
static char OPR_SYMBOL[] = 
  {' ',' ','-','+','+','-','M','E','*','.','C','A', 'J','<','O','`','\'','W'};
#endif


// Creates the structure polynomial through a series of calls to routines in 
// files Po_parse_poly.c, Po_expand_poly.c and Po_parse_exptext.c.
// First call Create_parse_tree() with In_str to get the parse tree 
// corresponding to In_str.                               
// Then expand the parse tree by calling Expand_parse_tree(). 
// Then get the simplified parse tree having basic operators like JUXT_PROD, 
// SCALAR_MULT along with small letters and scalars at the leaves by calling 
// Simplify_parse_tree(), in a loop until no further simplification is possible.
// The Simplified tree is translated into a string by calling Create_exp_str() 
// which traverses the tree in inorder.
// Finally the expanded string is parsed by calling Parse_exptext(), which 
// returns a pointer to the polynomial which in turn returned to the caller of 
// Create_poly().
// NOTE: The intermediate trees created during the creation of polynomial can 
//       be seen by switching on the DEBUG_EXP flag.
// Called from Main() in driver.c.
// arg: In_str -- polynomial string, operand portion of the commands 
//      "i [polynomial]" and "q [polynomial]".  
// RETURNS: Pointer to the created polynomial if In_str is successfully parsed,
//          NULL otherwise.                                         
polynomial* Create_poly (const char In_str[], size_t* In_str_len)
{
  static char** Exp_poly_str_ptr;  // Exp_poly_str may be expanded!

  bool Modified = 0;   // Has tree been changed in this call?
  char* exp_poly_str = (char*)Mymalloc (EXP_STR_LEN);
  size_t Maxsize_exp_poly_str = EXP_STR_LEN;
  exp_poly_str[0] = '\0';
  Exp_poly_str_ptr = &exp_poly_str;

  // Parse tree for the polynomial string In_str entered by user
  unexp_tnode* Unexp_tree1 = Create_parse_tree (In_str);
  if (Unexp_tree1 == NULL)
	{
	  return (NULL);
	}
#if DEBUG_EXP
  Print_tree (Unexp_tree1);
  std::cout << std::endl;
#endif

  // Expand the Parse tree by using formulae for operators
  unexp_tnode* Unexp_tree2 = Expand_parse_tree (Unexp_tree1); 
#if DEBUG_EXP
  Print_tree (Unexp_tree2);
  std::cout << std::endl;
#endif

  // Simplify the Expanded tree i.e. apply distributive laws.
  unexp_tnode* Exp_tree1 = Simplify_parse_tree (Unexp_tree2, &Modified); 
#if DEBUG_EXP
  unsigned int count = 1;
  std::cout << "simplified tree " << count << std::endl;
  Print_tree (Exp_tree1);
  std::cout << std::endl;
#endif

  unexp_tnode* Exp_tree2 = NULL;
  while (Modified) 
	{
	  Modified = 0;
	  Exp_tree2 = Simplify_parse_tree (Exp_tree1, &Modified); 
#if DEBUG_EXP
	  count++;
	  std::cout << "count = " << count << " Modified = " << Modified 
				<< std::endl << "simplified tree" << std::endl;
	  Print_tree(Exp_tree2);
	  std::cout << std::endl << "freeing Exp_tree1" << std::endl;
#endif
	  Free_tnode_tree (Exp_tree1);
	  Exp_tree1 = Exp_tree2;
	  Exp_tree2 = NULL;
    }

  // No more simplifications needed!
  Modified = 0;

  // Get rid of Unary minus by multiplying that subtree by -1
  unexp_tnode* Exp_tree3 = Elim_subtraction (Exp_tree1, &Modified); 
#if DEBUG_EXP
  count = 1;
  Print_tree (Exp_tree3);
  std::cout << std::endl;
#endif

  unexp_tnode* Exp_tree4 = NULL;
  while (Modified) 
	{
	  Modified = 0;
	  Exp_tree4 = Elim_subtraction (Exp_tree3, &Modified); 
	  
#if DEBUG_EXP
	  count++;
	  std::cout << "count = " << count << " Modified = " << Modified 
				<< std::endl;
	  Print_tree (Exp_tree4);
	  std::cout << std::endl << "freeing Exp_tree3" << std::endl;
#endif
	  Free_tnode_tree (Exp_tree3);
	  Exp_tree3 = Exp_tree4;
	  Exp_tree4 = NULL;
    }

  // Traverse the tree in inorder and write it as a string
  Create_exp_str (Exp_tree3, Exp_poly_str_ptr, &Maxsize_exp_poly_str);
  Free_tnode_tree (Unexp_tree1);
  Free_tnode_tree (Unexp_tree2);
  Free_tnode_tree (Exp_tree1);
  Free_tnode_tree (Exp_tree3);
  
#if DEBUG_EXP
  std::cout << exp_poly_str <<std::endl;
#endif

  // Parse the string to create the polynomial
  polynomial* poly = Parse_exptext (*Exp_poly_str_ptr);
  *In_str_len = strlen (exp_poly_str);
  free (exp_poly_str);
  return (poly);
} // end of Create_poly()





// Print the tree in preorder.
// Called only when DEBUG_EXP flag is on.
// Very useful to see how the parse tree gets simplified.
// arg: Pntr -- root of the tree of tnodes
#if DEBUG_EXP
void Print_tree (const unexp_tnode* Pntr)
{
  if (Pntr != NULL) 
	{
	  if ((Pntr->_operator != SMALL_LETTER)  && 
		  (Pntr->_operator != SCALAR)        &&
		  (Pntr->_operator != OPERATOR_PROD) && 
		  (Pntr->_operator != ARTIFICIAL_WORD)) 
		{
		  //printf ("%c(", OPR_SYMBOL[Pntr->_operator]);
		  std::cout << OPR_SYMBOL[Pntr->_operator] << "(";
		  Print_tree (Pntr->operand1);
		  if (Pntr->operand2 != NULL) 
			{
			  std::cout << ",";
			  Print_tree (Pntr->operand2);
			  if (Pntr->operand3 != NULL) 
				{
				  std::cout << ",";
				  Print_tree (Pntr->operand3);
				}
			}
		  std::cout << ")";
		}
	  else if (Pntr->_operator == SMALL_LETTER)
		{
		  std::cout << Pntr->s_letter;
		}
	  else if (Pntr->_operator == SCALAR)
		{
		  std::cout << Pntr->scalar_num;
		}
	  else if (Pntr->_operator == OPERATOR_PROD) 
		{
		  std::cout << OPR_SYMBOL[Pntr->_operator] << "(";
		  Print_tree (Pntr->operand1);
		  Print_opr_prod_tree (Pntr->operand2);
		  std::cout << ")";
        }
	  else if (Pntr->_operator == ARTIFICIAL_WORD) 
		{
		  std::cout << OPR_SYMBOL[Pntr->_operator] << "(",
		  Print_tree (Pntr->operand1);
		  Print_art_word (Pntr->operand2);
		  std::cout << ")";
        }
    }
} // end of Print_tree()



// Handle the special case of printing Operators.
// Print all operators in the list and their operands.
// NOTE: Called only by Print_tree(). i.e Called only for debugging.
// arg: 
// Pntr -- Pointer to a list of Operators (left, right) and their operands.
void Print_opr_prod_tree (const unexp_tnode* Pntr)
{
  while (Pntr != NULL) 
	{
	  std::cout << ",";
	  Print_tree (Pntr);
	  Pntr = Pntr->next;
    }
}



// Handle the special case of printing ARTIFICIAL_WORD tree.
// Print the tree corresponding to operator ARTIFICIAL_WORD.
// NOTE: Called only by Print_tree(). i.e Called only for debugging.
// arg: Pntr -- Pointer to a tree corresponding to operator ARTIFICIAL_WORD
void Print_art_word (const unexp_tnode* Pntr)
{
  while (Pntr != NULL) 
	{
	  std::cout << ",";
	  Print_tree (Pntr);
	  Pntr = Pntr->next;
    }
}
#endif // DEBUG_EXP




// Build the string pointed by Str_ptr, for the given tree, by traversing it in
// inorder.                                   
// NOTE: 
// Str_cat() is used for string concatenation. This special function is needed 
// because when size of the string pointed by Str_ptr may not be enough to be 
// concatenated with another string. In which case Str_cat() reallocates more 
// space for the string pointed by Str_ptr and changes the Max_size to reflect 
// the change in string size. Thus string grows varies dynamically to 
// accommodate large polynomials.
// MODIFIES: Str_ptr -- Pointer to the string
//           Maxsize_ptr -- Maxsize of the string pointed by Str_ptr
// arg: Pntr -- pointer to the root of the expanded polynomial tree
void Create_exp_str 
(const unexp_tnode* Pntr, char** Str_ptr, size_t* Maxsize_str_ptr)
{
  char Temp_str[10];
  if (Pntr != NULL) 
	{
	  if ((Pntr->_operator != SMALL_LETTER) && (Pntr->_operator != SCALAR)) 
		{
		  if (Pntr->_operator == JUXT_PROD) 
			{
			  sprintf (Temp_str, "(");
			  Str_cat (Str_ptr, Temp_str, Maxsize_str_ptr);
            }

		  Create_exp_str (Pntr->operand1, Str_ptr, Maxsize_str_ptr);

		  if (Pntr->_operator == ADDITION) 
			{
			  if (! Found_minus (Pntr->operand2)) 
				{
				  sprintf (Temp_str, "+");
				  Str_cat (Str_ptr, Temp_str, Maxsize_str_ptr);
                }
            }
		  else if (Pntr->_operator == SUBTRACTION) 
			{
			  sprintf (Temp_str, "-");
			  Str_cat (Str_ptr, Temp_str, Maxsize_str_ptr);
            }

		  Create_exp_str (Pntr->operand2, Str_ptr, Maxsize_str_ptr);
		  
		  if (Pntr->_operator == JUXT_PROD) 
			{
			  sprintf (Temp_str, ")");
			  Str_cat (Str_ptr, Temp_str, Maxsize_str_ptr);
            }
        }
	  else if (Pntr->_operator == SMALL_LETTER) 
		{
		  sprintf (Temp_str, "%c", Pntr->s_letter);
		  Str_cat (Str_ptr, Temp_str, Maxsize_str_ptr);
		}
	  else if (Pntr->_operator == SCALAR) 
		{
		  if (Pntr->scalar_num == -1) 
			{
			  sprintf (Temp_str, "-");
			  Str_cat (Str_ptr, Temp_str, Maxsize_str_ptr);
            }
		  else if (Pntr->scalar_num != 1) 
			{
			  sprintf (Temp_str, "%d", Pntr->scalar_num);
			  Str_cat (Str_ptr, Temp_str, Maxsize_str_ptr);
            }
        }
    }
} // end of Create_exp_str()





// Since the tree is traversed in inorder, avoid printing something like x+-4y.
// i.e. + needs to be suppressed in case of scalars less than 0, thus correctly
// prints x-4y.  
// arg: Pntr -- Pointer to a tree
// RETURNS: 1 if first if any of the operand1's of the tree are scalars less 
//          than 0, 0  otherwise.
bool Found_minus (const unexp_tnode* Pntr)
{
  if (Pntr == NULL)
	{
	  return (0);
	}
  else if (Pntr->scalar_num < 0)
	{
	  return (1);
	}
  else 
	{
	  return (Found_minus (Pntr->operand1));
	}
}




// Free all the space used for the tree with root pointer as Pntr.
// Traveses the tree in inorder to free all the nodes. The nodes are freed and 
// entered onto Free tnode queue through a call to Free_tnode().
// MODIFIES: Pntr -- Pointer to a tree
void Free_tnode_tree (unexp_tnode* Pntr)
{
  if (Pntr != NULL) 
	{
	  if ((Pntr->operand1 == NULL) && (Pntr->operand2 == NULL) &&
		  (Pntr->operand3 == NULL) && (Pntr->next == NULL))
		{
		  Free_tnode (Pntr);
		}
	  else 
		{
		  Free_tnode_tree (Pntr->operand1);
		  Free_tnode_tree (Pntr->operand2);
		  Free_tnode_tree (Pntr->operand3);
		  Free_tnode_tree (Pntr->next);
		  Free_tnode (Pntr);
		}
	}
}




// TODO: eliminate this altogether
/* MODIFIES:                                                       */
/*     Str1_ptr -- pointer to a string                             */
/*     *Str1_maxsize_ptr -- maximum size to which Str1 can grow.   */ 
/* REQUIRES:                                                       */
/*     Str2 -- string which is to be attached to *Str1_ptr.        */
/* FUNCTION:                                                       */
/*     Str2 is concatenated to Str1.                               */
/* NOTE:                                                           */
/*     If Str1 is not big enough to be concatenated with str2,     */
/*     extra space is allocated, *str1_ptr is changed to point to  */
/*     Str1 whose size has expanded and *Str1_maxsize_ptr is       */
/*     changed to reflect the increased size of Str1.              */
/*     If unable to find space, No_memory_panic() will be called.  */
static void Str_cat (char** Str1_ptr, const char Str2[], size_t* Str1_maxsize_ptr)
{
  if ((*Str1_maxsize_ptr) < (strlen(*Str1_ptr) + strlen(Str2)))
    {
	  size_t new_size = 2 * ((*Str1_maxsize_ptr) + strlen(Str2));
	  *Str1_ptr = (char*)realloc (*Str1_ptr, new_size);
      if (*Str1_ptr == NULL)
		{
		  No_memory_panic();
		}
      else 
		{
		  strcat (*Str1_ptr, Str2);
		  *Str1_maxsize_ptr = new_size;
		}
    }
  else
	{
	  strcat (*Str1_ptr, Str2);
	}
}


// end of Po_create_poly.cc

