/*
  file:         albert.cc
  purpose:      Albert class
  created:      pasha mar 10 2009
  modified:     pasha jan 1-4 2019
  modification: - eliminated MAX_LINE
                - some char*'s replaced by std::string
				- open .albert with ifstream and explicitly close at the end
*/

#include <gmpxx.h>
#include <string.h>
#include <unistd.h>
#include <fstream>
#include "albert.hh"
#include "GF2.hh"
#include "GFp.hh"

#ifdef DMALLOC
#include <dmalloc.h>
#endif

// built-in modular arithemtic
unsigned long long Albert::PRIME;
unsigned long long GFp_element::HALF_PRIME;
std::map<unsigned long long, long long> GFp_element::Inverse_table;
std::map<unsigned long long, long long>::const_iterator GFp_element::ITI;



Albert::Albert (char* albert_dot_file, const bool verbose)
{
  this->block_list_head = NULL;
  this->current_block   = NULL;
  this->mtable_present  = 0;
  this->mult_table      = NULL;
  this->node_list_head  = NULL;
  this->status_GenerateEquations = 1;
  this->status_Multpart          = 1;
  this->status_PerformSubs       = 1;

  this->field_elem_type = INVALID_TYPE;
  this->set_char (DEFAULT_CHAR, 0);

  this->Id_queue.first = NULL;
  

  // Fill in Dalbertheadptr args with the structure from .albert file, 
  // (NULL if .albert not present or unable to read).
  // Reads the .albert file if present and forms a list of definitions.
  // Linked list representation means there could be any number of definitions 
  // in .albert.
  // A definition can start anywhere on a line.
  // Blank lines are permitted anywhere in .albert.
  // Line after % is totally ignored.
  // A definition can be continued on the next line by ending it with a /.
  // Comments can appear between two lines of a definition or at the of any 
  // line.
  // A blank ' ' or a '\t' seperates the left hand side and right hand side of a
  // definition.
  // If Albert dumps the core during initialization, probably there is a bug in
  // this code. Albert can be used either by changing the .albert file or by 
  // getting rid of it completely.

  this->Dalbertheadptr = NULL;
  char cwd[200];
  bool inCWD;		// flag to determine if file is in cwd
  if (! strncmp (".albert", albert_dot_file, strlen (albert_dot_file)))
    {
      getcwd (cwd, 200);
      strcpy (albert_dot_file, cwd);
      strcat (albert_dot_file, "/.albert");
      inCWD = 1;
    }
  else
    {
      getcwd (cwd, 200);
      inCWD = 0;
    }

  std::ifstream albert_fp;
  albert_fp.open (albert_dot_file);
  if (! albert_fp.is_open())
    {
      // check the specified path 1st
      if (! inCWD)
		{
		  if (verbose)
			{
			  std::cout << "Can't open " << albert_dot_file << " file." 
						<< std::endl
						<< "Attempting to read " << cwd << "/.albert" 
						<< std::endl;
			}
		  albert_fp.open (".albert");
		  if (! albert_fp.is_open())
			{
			  if (verbose)
				{
				  std::cout << "Can't open " << cwd << "/.albert file." 
							<< std::endl;
				}
			  return;
			}
		  else
			{
			  if (verbose)
				{
				  std::cout << "Using " << cwd << "/.albert" << std::endl;
				}
			}
		}
      else
		{
		  if (verbose)
			{
			  std::cout << "Can't open " << cwd << "/.albert file." 
						<< std::endl;
			}
		  return;
		}
    }
  else if (verbose)
    {
      std::cout << "Using " << albert_dot_file << std::endl;
    }

  // Head to the list of .albert defs
  this->Dalbertheadptr = (Dalbert_node*)Mymalloc (sizeof(Dalbert_node));
  this->Dalbertheadptr->lhs  = NULL;
  this->Dalbertheadptr->rhs  = NULL;
  this->Dalbertheadptr->next = NULL;
  Dalbert_node* tnodeptr = this->Dalbertheadptr;

  std::string albert_line;
  std::string temp_line;
  std::string temp_line2;
  bool temp;  // flag to indicate wheter eof occured
  int temp_line_len, lhs_len, rhs_len, alinelen, lhsstart;
  int i=0;
  int j=0;
  bool slashfound = 1;
  bool blankfound = 1;
  while (std::getline (albert_fp, temp_line), temp = albert_fp.good())
	{
	  slashfound = 1;
	  while (slashfound && temp)
		{
		  blankfound = 1;
		  while (blankfound && temp)
			{
			  temp_line_len = temp_line.length();
			  for (i=0; i<temp_line_len; i++) 
				{
				  if (temp_line[i] == '%') 
					{
					  temp_line[i] = '\0';
					  break;
					}
				}
			  temp_line_len = temp_line.length();
			  for (i=0; i<temp_line_len; i++) 
				{
				  if ((temp_line[i] != ' ') && (temp_line[i] != '\t') &&
					  (temp_line[i] != '\n'))
					{
					  blankfound = 0;
					}
				}
			  if (blankfound)
				{
				  std::getline (albert_fp, temp_line);
				  temp = albert_fp.good();
				}
			}
		  i = 0;
		  temp_line2.resize (0);
		  while ((temp_line[i] != '\\') && (temp_line[i] != '\0') &&
				 (temp_line[i] != '%') && (temp_line[i] != '\n'))
			{
			  temp_line2.push_back (temp_line[i++]);
			}
		  if (temp_line[i] == '\\') 
			{
			  slashfound = 1;
			  std::getline (albert_fp, temp_line); temp = albert_fp.good();
			}
		  else
			{
			  slashfound = 0;
			}
		  albert_line += temp_line2;
		}
	  
	  i = 0;
	  albert_line.push_back ('\0');
	  alinelen = albert_line.length();
	  while (i < alinelen) 
		{
		  if ((albert_line[i] != ' ') && (albert_line[i] != '\t'))
			{
			  break;
			}
		  i++;
	  }
	  lhs_len = 0;
	  lhsstart = i;
	  while (i < alinelen) 
		{
		  if ((albert_line[i] == ' ') || (albert_line[i] == '\t'))
			{
			  break;
			}
		  lhs_len++;
		  i++;
		}
	  rhs_len = 0;
	  while (i < alinelen) 
		{
		  if ((albert_line[i] != ' ') && (albert_line[i] != '\t'))
			{
			  rhs_len++;
			}
		  i++;
		}
	  if ((lhs_len > 0) && (rhs_len > 0))  
		{
		  tnodeptr->lhs = (char*)Mymalloc (lhs_len+1);
		  i = lhsstart;
		  j = 0;
		  while (i < alinelen) 
			{
			  if ((albert_line[i] == ' ') || (albert_line[i] == '\t'))
				{
				  break;
				}
			  tnodeptr->lhs[j++] = albert_line[i++];
			}
		  tnodeptr->lhs[j++] = '\0';
		
		  tnodeptr->rhs = (char*)Mymalloc (rhs_len+1);
		  j = 0;
		  while (i < alinelen) 
			{
			  if ((albert_line[i] != ' ') && (albert_line[i] != '\t'))
				{
				  tnodeptr->rhs[j++] = albert_line[i];
				}
			  i++;
			}
		  tnodeptr->next = (Dalbert_node*)Mymalloc (sizeof(Dalbert_node));
		  tnodeptr = tnodeptr->next;
		  tnodeptr->lhs = NULL;
		  tnodeptr->rhs = NULL;
		  tnodeptr->next = NULL;
		}
	  albert_line.resize (0);
	} // end of loop thru lines of .albert file

  if (albert_fp.is_open())
	{
	  albert_fp.close();
	}

#if DB_READ_ALBERT
  i = 1;
  tnodeptr = this->Dalbertheadptr;
  while (tnodeptr->next != NULL) 
    {
      std::cout << " " << i++ << " " 
				<< strlen(tnodeptr->lhs) << "  " << tnodeptr->lhs << " " 
				<< strlen(tnodeptr->rhs) << "  " << tnodeptr->rhs << " " 
				<< std::endl;
      tnodeptr = tnodeptr->next;
    }
#endif
  
} // end of Albert ctor




Albert::~Albert()
{
  // free the linked list structure containing .albert info
  Dalbert_node* current;
  while (this->Dalbertheadptr != NULL)
    {
      current = this->Dalbertheadptr;
      free (current->lhs);
      free (current->rhs);
      this->Dalbertheadptr = current->next;
      free (current);
    }
  
  this->destroy_mult_table();
} // end of Albert dtor





////////////////////////////////////////////////////////////////////////////
//                       clear_mult_table                                 //
////////////////////////////////////////////////////////////////////////////
void Albert::clear_mult_table()
{
  switch (this->field_elem_type)
	{
	case GF2_TYPE:
	  this->clear_mult_table<GF2_element>();
	  break;
	  
	case GFP_TYPE:
	  this->clear_mult_table<GFp_element>();
	  break;
	  
	case GMP_TYPE:
	  this->clear_mult_table<mpq_class>();
	  break;
	  
	default:
	  std::cerr << "wrong field_elem_type=" << this->field_elem_type
				<< std::endl;
	  break;
	}
  if (this->mtable_present) 
	{
	  this->mtable_present = 0;
	  std::cout << "Destroyed the Multiplication Table." << std::endl;
	}
} // end of Albert::clear_mult_table()





////////////////////////////////////////////////////////////////////////////
//                       destroy_mult_table                               //
////////////////////////////////////////////////////////////////////////////
void Albert::destroy_mult_table()
{
  switch (this->field_elem_type)
	{
	case GF2_TYPE:
	  this->DestroyMultTable<GF2_element>();
	  break;
		  
	case GFP_TYPE:
	  this->DestroyMultTable<GFp_element>();
	  break;
	  
	case GMP_TYPE:
	  this->DestroyMultTable<mpq_class>();
	  break;
	  
	case INVALID_TYPE:
	  return;

	default:
	  std::cerr << "wrong field_elem_type=" << this->field_elem_type
				<< std::endl;
	  break;
	}
  this->mult_table = NULL;
  if (this->mtable_present) 
	{
	  this->mtable_present = 0;
	  std::cout << "Destroyed the Multiplication Table." << std::endl;
	}
} // end of Albert::destroy_mult_table()





////////////////////////////////////////////////////////////////////////////
//                       print_mtable                                     //
////////////////////////////////////////////////////////////////////////////
void Albert::print_mtable (std::ostream& os, const bool output_for_gap) const
{
  switch (this->field_elem_type)
	{
	case GF2_TYPE:
	  this->Print_MultTable<GF2_element> (os, output_for_gap);
	  break;

	case GFP_TYPE:
	  this->Print_MultTable<GFp_element> (os, output_for_gap);
	  break;
	  
	case GMP_TYPE:
	  this->Print_MultTable<mpq_class> (os, output_for_gap);
	  break;

	default:
	  std::cerr << "wrong field_elem_type=" << this->field_elem_type
				<< std::endl;
	  break;
	}
} // end of Albert::print_mtable()





////////////////////////////////////////////////////////////////////////////
//                       set_char                                         //
////////////////////////////////////////////////////////////////////////////
// set charactersitic of the ground field
// returns 1 on success, 0 on failure
bool Albert::set_char (const long long& p, const bool verbose)
{
  FieldElemType new_field_elem_type = INVALID_TYPE;

  if (p == 0)
    {
      // rational arithmetic via GMP
	  new_field_elem_type = GMP_TYPE;
    }
  else if (p < 2)
	{
	  std::cerr << "wrong value p=" << p << ". Characteristic not changed."
				<< std::endl;
	  return (0);
	}
  else if (p == 2)
	{
	  new_field_elem_type = GF2_TYPE;
	}
  else
	{
	  // check that p is a prime
	  char str[20];
	  sprintf (str, "%lld", p);
	  mpz_t I;
	  mpz_init (I);
	  mpz_set_str (I, str, 10);
	  int r = mpz_probab_prime_p (I, 50);
	  mpz_clear (I);
	  switch (r)
		{
		case 0:
		  std::cout << p << " is not prime" << std::endl;
		  return (0);
		case 1:
		  // p is probably prime
		  std::cout << p << " is probably prime" << std::endl;
		  break;
		case 2:
		  // p is prime
		  break;
		default:
		  std::cerr << "wrong return value " << r 
					<< " when testing for primeness" << std::endl;
		}

	  // built-in modular arithmetic
	  new_field_elem_type = GFP_TYPE;
	  GFp_element::HALF_PRIME = (p - 1) >> 1;
	  GFp_element::Inverse_table.clear();
	}

  if (this->field_elem_type != new_field_elem_type)
	{
	  this->destroy_mult_table();
	  this->field_elem_type = new_field_elem_type;
	}
  if (Albert::PRIME != (unsigned long long)p)
	{
	  this->clear_mult_table();
	  Albert::PRIME = p;
	  if (verbose)
		{
		  std::cout << "Characteristic changed to p = " << p << std::endl;
		}
	}

  return (1);
} // end of Albert::set_char()


// end of albert.cc
