/*
  file:         driver.cc
  purpose:      This is the Main module of Albert. It does initializations.
                Interprets and executes Commands. Gets Command line and executes
                the command.
  created:      Sekhar Muddana circa 1990
  modified:     pasha dec 30 2018-jan 1,4,10 2019
  modification: ver. 4.0M6: 
                - minor code reshuffling
				- some int -> size_t
				- few char*'s converted to std::string; eliminated COMMAND_LEN
*/

// TODO: convert all char* variables to std::string

#include <gmpxx.h>
#include <stdio.h>   // needed for GNU readline
#include <readline/readline.h>
#include <readline/history.h>
#include <fstream>
#include "albert.hh"
#include "dotalbert.hh"
#include "GF2.hh"
#include "GFp.hh"
#include "Ty_routines.hh"

#ifdef DMALLOC
#include <dmalloc.h>
#endif

static size_t __getline (std::string&);
static void usage();
static bool Compatible (const polynomial*, const P_type);
static bool Substr (const std::string&, const std::string&);

bool problem_type_present = 0;



int main (int argc, char* argv[])
{
  // BEGIN argument handling
  if (argc > 3)
	{
	  usage();
	  return (-1);
	}

  char dir[100];  // location of .albert
  dir[0] = '\0';

  int argPos, i;
  for (i=1; i<argc; ++i)
	{
	  if (argv[i][0] != '-')
		{
		  // make sure it's a flag 
		  std::cout << "flag, " << argv[i] << ", requires a \"-\"." 
					<< std::endl;
		  usage();
		}
	  else
		{
		  if (strlen (argv[i]) > 2)
			{
			  argPos = 2;
			}
		  else
			{
			  argPos = 0;
			}
		  switch (argv[i][1])
			{             
			  // get the flag parameters 
			case 'a':            // .albert directory
			  if (argPos)
				{
				  strcpy (dir, argv[i] + argPos);
				}
			  else
				{
				  ++i;
				  strcpy (dir, argv[i]);
				}
			  if (dir[strlen(dir)-1] != '/')
				{
				  strcat (dir, "/");
				}
			  strcat (dir, ".albert");
			  break;
			  
			default:
			  std::cout << argv[i][1] << " is an invalid flag type." 
						<< std::endl;
			  usage();
			}
		}
	} // end of loop thru command line
  
  if (! strlen (dir))
	{
	  strcpy (dir, ".albert");
	}
  // end of argument handling

  std::cout << std::endl 
			<< "Albert version 4.0M6, 2019"
			<< std::endl << std::endl;

  Albert A (dir, 1);

  // initialize the problem type
  P_type ptype;
  for (i=0; i<NUM_LETTERS; i++)
    {
      ptype.degrees[i] = 0;
    }
  ptype.tot_degree = 0;

  // support for save and view commands
  char tableFileName[100];
  char table;
  extern Deg_to_basis_rec* Deg_to_basis_table;
  std::filebuf tableFilePtr;

  std::string Command, Operand;
  polynomial* Cur_poly;
  P_type Temp_type;   // structure for problem type
  char c;
  long long int p;
  std::string temp_line;
  std::string temp_line2;
  int Cur_id_no, temp_line_len, j, tl;
  bool first_generator, match, slashfound;
  std::string total_line;
  char *str1, *str2, *str3;
  Dalbert_node* tnodeptr;

  rl_bind_key ('\t', rl_insert);  // disable default tab completion in readline

  bool quit = 0;
  while (! quit)
    {
      std::cout << std::endl;
	  
      ////////// get the command and the operand ///////////////////////
      //
	  // Also uses the definitions in .albert
	  //
	  total_line.resize (0);
      temp_line_len = __getline (temp_line);
	  temp_line2.resize (0);

      slashfound = 1;
      while ((slashfound) && (temp_line_len > 0)) 
		{
		  i = 0;
		  while (temp_line[i] == ' ')
			{ 
			  // read past leading white space
			  ++i;
			}
		  while ((temp_line[i] != '\\') && (temp_line[i] != '\0'))
			{
			  temp_line2.push_back (temp_line[i++]);
			}
		  if (temp_line[i] == '\\') 
			{
			  slashfound = 1;
			  temp_line_len = __getline (temp_line);
			}
		  else
			{
			  slashfound = 0;
			}
		  temp_line2.push_back ('\0');
		  total_line += temp_line2;
		}
      
      if (temp_line_len == -1)
		{
		  // error reading stdin, abort
		  goto END;
		}
	  
      i = 0;
	  tl = total_line.length();
      char* total_line1 = (char*)Mymalloc (tl+1);
      while (i < tl)
		{
		  if ((total_line[i] == ' ') || (total_line[i] == '\t'))
			{
			  break;
			}
		  total_line1[i] = total_line[i];
		  i++;
		}
      total_line1[i++] = '\0';

	  Command = total_line1;
      j = 0;
      while (i < tl) 
		{
		  if ((total_line[i] != ' ') && (total_line[i] != '\t'))
			{
			  total_line1[j++] = total_line[i++];
			}
		  else
			{
			  i++;
			}
		}
      total_line1[j] = '\0';
	  total_line = total_line1;
	  
      if (A.Dalbertheadptr != NULL)
		{
		  while (1) 
			{
			  // Scans the total_line and breaks it into 3 strings str1, str2 
			  // and str3. str1 is the prefix before $ in total_line. str2 is 
			  // the string between 2 $'s. str3 is the tail. Now str2 is 
			  // substituted with a string as per the definitions in .albert. 
			  // i.e if str2 matches with some left hand side of a definition in
			  // .albert, then it is replaced with corresponding right hand 
			  // side.
			  tnodeptr = A.Dalbertheadptr;
			  tl = total_line.length();
			  if (tl < 3)
				{
				  break;
				}
			  str1 = (char*)Mymalloc (tl+1);
			  i = 0;
			  j = 0;
			  while ((i<tl) && (total_line[i] != '$'))
				{
				  str1[j++] = total_line[i++];
				}
			  str1[j] = '\0';
			  if (total_line[i++] != '$') 
				{
				  free (str1);
				  break;
				}
			  str2 = (char*)Mymalloc (tl+1);
			  j = 0;
			  while ((i<tl) && (total_line[i] != '$'))
				{
				  str2[j++] = total_line[i++];
				}
			  str2[j] = '\0';
			  if (total_line[i++] != '$') 
				{
				  free (str1);
				  free (str2);
				  break;
				}
			  str3 = (char*)Mymalloc (tl+1);
			  j = 0;
			  while (i<tl)
				{
				  str3[j++] = total_line[i++];
				}
			  str3[j] = '\0';
			  
			  match = 0;
			  while ((tnodeptr->next != NULL) && (! match))
				{
				  if (strcmp (tnodeptr->lhs, str2) == 0) 
					{
					  match = 1;
					  break;
					}
				  tnodeptr = tnodeptr->next;
				}
			  if (! match) 
				{
				  free (str1);
				  free (str2);
				  free (str3);
				  break;
				}
			  total_line.resize (0);
			  total_line += str1;
			  total_line += tnodeptr->rhs;
			  total_line += str3;
#if DB_READ_ALBERT
			  std::cout << "str1 = " << str1 << std::endl 
						<< "str2 = " << str2 << std::endl
						<< "str3 = " << str3 << std::endl
						<< "total line = " << total_line << std::endl;
#endif
			  free (str1);
			  free (str2);
			  free (str3);
    	    } // end of substitution loop
		}
	  
	  Operand = total_line;
#if DB_READ_ALBERT
      std::cout << "command = " << Command << std::endl
				<< "operand = " << Operand << std::endl;
#endif
      free (total_line1);
      ////////// end of get the command and the operand ////////////////
      
	  
      std::cout << std::endl;
      switch (Command[0]) 
		{
		case 'i':
		  // "identity polynomial". Accept new identity.
		  if (! Substr (Command, "identity")) 
			{
			  std::cout << "Illegal command.";
			  break;
			}
		  Cur_id_no = A.add_identity (Operand.c_str());
		  if (Cur_id_no > 0)
			{
			  Print_poly (A.Cur_id, A.Cur_poly_len);
			  std::cout << std::endl << "Entered as identity " << Cur_id_no 
						<< "." << std::endl;
			}
		  break;
		  
		case 'r':
		  // "remove number|*". remove identity.
		  if (! Substr (Command, "remove")) 
			{
			  std::cout << "Illegal command.";
			  break;
			}
		  if (Operand[0] == '*') 
			{
			  Remove_all_ids (&A.Id_queue);
			  std::cout << "Removed all the identities." << std::endl;
			  A.clear_mult_table();
			}
		  else 
			{
			  Cur_id_no = atoi (Operand.c_str());
			  if (Cur_id_no > 0) 
				{
				  if (Remove_id (Cur_id_no, &A.Id_queue)) 
					{
					  std::cout << "Identity " << Cur_id_no << " removed." 
								<< std::endl;
					  A.clear_mult_table();
					}
				  else
					{
					  std::cout << "Identity " << Cur_id_no << " not present.";
					}
				}
			  else
				{
				  std::cout << "Number must be at least 1.";
				}
			}
		  break;
		  
		case 'g':
		  // "generators word". Accept problem type.
		  if (! Substr (Command, "generators")) 
			{
			  std::cout << "Illegal command." << std::endl;
			  break;
			}
		  if (! Parse_generator_word (Operand.c_str(), &Temp_type))
			{
			  std::cout << "Error in the generator list." << std::endl;
			}
		  else if (Temp_type.tot_degree < 2)
			{
			  std::cout << "Total degree must be at least 2." << std::endl;
			}
		  else 
			{
			  for (i=0; i<NUM_LETTERS; i++)
				{
				  ptype.degrees[i] = 0;
				}
			  ptype.tot_degree = 0;
			  
			  for (i=0; i<NUM_LETTERS; i++)
				{
				  ptype.degrees[i] = Temp_type.degrees[i];
				}
			  ptype.tot_degree = Temp_type.tot_degree;
			  if (problem_type_present) 
				{
				  std::cout << "Problem type changed." << std::endl;
				}
			  else
				{
				  std::cout << "Problem type stored." << std::endl;
				}
			  AssignBasisNumberstoLetters (ptype);

			  // struct P_type ptype is used to create Target type Type 
			  // problem_target_type
			  // If generators are 3a2b2c then Target type is 322.
			  unsigned int ttype_len = 1;
			  for (i=0; i<NUM_LETTERS; i++)
				{
				  if (ptype.degrees[i] > 0)
					{
					  ttype_len++;
					}
				}
			  A.Target_type.resize (ttype_len);
			  unsigned int cur_tt_index = 0;
			  for (i=0; i<NUM_LETTERS; i++)
				{
				  if (ptype.degrees[i] > 0)
					{
					  A.Target_type[cur_tt_index++] = ptype.degrees[i];
					}
				}
			  A.Target_type[cur_tt_index] = 0;
			  problem_type_present = 1;
			  A.clear_mult_table();
			}
		  break;
		  
		case 'f':
		  // "field number". Change the charactersitic of the ground field.
		  if (! Substr (Command, "field")) 
			{
			  std::cout << "Illegal command.";
			  break;
			}
		  p = atoll (Operand.c_str());
		  A.set_char (p, 1);
		  break;
		  
		case 'b':
		  // "build". Build the multiplication table.
		  if (! Substr (Command, "build")) 
			{
			  std::cout << "Illegal command.";
			  break;
			}
		  if (! problem_type_present) 
			{
			  std::cout << "Problem Type not present." << std::endl 
						<< "Can't build Multiplication Table." << std::endl;
			  break;
			}
		  switch (A.field_elem_type)
			{
			case GF2_TYPE:
			  A.build<GF2_element> (1);
			  break;

			case GFP_TYPE:
			  A.build<GFp_element> (1);
			  break;

			case GMP_TYPE:
			  A.build<mpq_class> (1);
			  break;

			default:
			  std::cerr << "wrong field_elem_type=" << A.field_elem_type
						<< std::endl;
			  break;
			}
		  break;
		  
		case 'd':
		  // "display". Display the current problem configuration.
		  if (! Substr (Command, "display")) 
			{
			  std::cout << "Illegal command.";
			  break;
			}
		  if (A.Id_queue.first != NULL) 
			{
			  std::cout << "Defining identities are:" << std::endl;
			  Print_ids (A.Id_queue);
			}
		  else
			{
			  std::cout << "No defining identities." << std::endl;
			}
		  std::cout << "Characteristic p = " << Albert::PRIME << "." 
					<< std::endl;
		  
		  // Print the Problem type in a good looking format!
		  if (problem_type_present) 
			{
			  first_generator = 1;
			  std::cout << "Problem type = [";
			  for (i=0; i<NUM_LETTERS; i++) 
				{
				  if (ptype.degrees[i] == 1) 
					{
					  c = 'a' + i;
					  if (! first_generator)
						{
						  std::cout << "," << c;
						}
					  else 
						{
						  first_generator = 0;
						  std::cout << c;
						}
					}
				  else if (ptype.degrees[i] > 1) 
					{
					  c = 'a' + i;
					  if (! first_generator)
						{
						  std::cout << "," << (unsigned short)ptype.degrees[i] << c;
						}
					  else 
						{
						  first_generator = 0;
						  std::cout << (unsigned short)ptype.degrees[i] << c;
						}
					}
				}
			  std::cout << "]; Total degree = " 
						<< (unsigned int)ptype.tot_degree << "." << std::endl;
			}
		  else
			{
			  std::cout << "Problem type not specified." << std::endl;
			}
		  
		  std::cout << "Multiplication table ";
		  if (! A.mtable_present)
			{
			  std::cout << "not ";
			}
		  std::cout << "present.";
		  break;
		  
		case 'p':
		  // "polynomial poly". Enquire if poly is an identity
		  if (! Substr (Command, "polynomial")) 
			{
			  std::cout << "Illegal command.";
			  break;
			}
		  if (! A.mtable_present) 
			{
			  std::cout << "Multiplication table not present.";
			  break;
			}
		  Cur_poly = Create_poly (Operand.c_str(), &A.Cur_poly_len);
		  if (Cur_poly == NULL) 
			{
			  std::cout << Operand << " is not a polynomial.";
			  break;
			}
		  if (! Homogeneous (Cur_poly)) 
			{
			  std::cout << "Given polynomial is not homogeneous.";
			  break;
			}
		  if (Cur_poly->degree < 2) 
			{
			  std::cout << "Total degree must be at least 2.";
			  break;
			}
		  if (! Compatible (Cur_poly, ptype)) 
			{
			  std::cout << "Polynomial type not a subtype of Target_type.";
			  break;
			}

		  switch (A.field_elem_type)
			{
			case GF2_TYPE:
			  if (A.IsIdentity<GF2_element> (Cur_poly))
				{
				  std::cout << "Polynomial is an identity.";
				}
			  else
				{
				  std::cout << "Polynomial is not an identity.";
				}
			  break;
			  
			case GFP_TYPE:
			  if (A.IsIdentity<GFp_element> (Cur_poly))
				{
				  std::cout << "Polynomial is an identity.";
				}
			  else
				{
				  std::cout << "Polynomial is not an identity.";
				}
			  break;
			  
			case GMP_TYPE:
			  if (A.IsIdentity<mpq_class> (Cur_poly))
				{
				  std::cout << "Polynomial is an identity.";
				}
			  else
				{
				  std::cout << "Polynomial is not an identity.";
				}

			  break;
			  
			default:
			  std::cerr << "impossible field element type " 
						<< A.field_elem_type 
						<< std::endl;
			  break;
			}
		
		  DestroyPoly (Cur_poly);
		  break;
		  
		case 'x':
		  // "xpand poly"
		  if (! Substr (Command, "xpand")) 
			{
			  std::cout << "Illegal command.";
			  break;
			}
		  Cur_poly = Create_poly (Operand.c_str(), &A.Cur_poly_len);
		  if (Cur_poly != NULL) 
			{
			  if (! Homogeneous (Cur_poly)) 
				{
				  std::cout << "Given polynomial is not homogeneous.";
				  DestroyPoly (Cur_poly);
				  break;
				}
			  else 
				{
				  Print_poly (Cur_poly, A.Cur_poly_len);
				  std::cout << std::endl;
				  DestroyPoly (Cur_poly);
				}
			}
		  else 
			{
			  std::cout << Operand << " is not a polynomial.";
			}
		  break;
		  
		  // "save" the multiplication table or the basis table to a file
		case 's':
		  if (! Substr (Command, "save"))
			{
			  std::cout << "Illegal command.";
			  break;
			}
	  
		  table = Operand[0];
		  switch (table) 
			{
			case 'b':
			  if (A.mtable_present && (A.Basis_table.size() > 1) && 
				  Deg_to_basis_table)
				{
				  std::cout << "File Name --> ";
				  fflush (stdout);
				  fgets (tableFileName, sizeof(tableFileName), stdin);
				  tableFileName[ strlen(tableFileName) -1 ] = 0;
				  std::cout << std::endl;
				  if (strlen (tableFileName))
					{
					  if (! tableFilePtr.open (tableFileName, std::ios::out))
						{
						  std::cout << "Unable to open file " 
									<< tableFileName << std::endl;
						  break;
						}
					}
				  else
					{
					  std::cout << "No file name was entered. Command aborted"
								<< std::endl;
					  break;
					}
				  std::ostream os (&tableFilePtr);
				  A.PrintBasisTable (os);
				  tableFilePtr.close();
				}
			  else
				{
				  std::cout << "Basis Table not present." << std::endl;
				}
			  break;
			  
			case 'm':
			  if (A.mtable_present)
				{
				  std::cout << "File Name --> ";
				  fflush (stdout);
				  fgets (tableFileName, sizeof(tableFileName), stdin);
				  tableFileName[strlen(tableFileName) - 1] = 0;
				  std::cout << std::endl;
				  if (strlen (tableFileName))
					{
					  if (! tableFilePtr.open (tableFileName, std::ios::out))
						{
						  std::cout << "Unable to open file "
									<< tableFileName << std::endl;
						  break;
						}
					}
				  else
					{
					  std::cout << "No file name was entered. Command aborted"
								<< std::endl;
					  break;
					}
				  std::ostream os (&tableFilePtr);
				  A.print_mtable (os, strstr (Operand.c_str(), "gap") != NULL);
				  tableFilePtr.close();
				}
			  else
				{
				  std::cout << "Multiplication Table not present" << std::endl;
				}
			  break;
			  
			default:
			  std::cout << "Invalid table type. Specify \"m\" or \"b\"" 
						<< std::endl;
			}
		  break;
		  
		  // "view" the multiplication table or the basis table on screen
		case 'v':
		  if (! Substr (Command, "view"))
			{
			  std::cout << "Illegal command.";
			  break;
			}
		  
		  table = Operand[0];
		  switch (table)
			{
			case 'b':
			  if (A.mtable_present && (A.Basis_table.size() > 1) && 
				  Deg_to_basis_table)
				{
				  A.PrintBasisTable (std::cout);
				}
			  else
				{
				  std::cout << "Basis Table not present." << std::endl;
				}
			break;
	    
		  case 'm':
			if (A.mtable_present)
			  {
				A.print_mtable (std::cout, strstr (Operand.c_str(), "gap") != NULL);
			  }
			else
			  {
				std::cout << "Multiplication Table not present." << std::endl;
			  }
			break;
			
		  default:
			std::cout << "Invalid table type " << table << "." << std::endl
					  << "Table type must be \"m\" or \"b\"" << std::endl;
		  }
		  break;
	  
		case 't':
		  // "type nonassoc. word". Find the type of nonassoc. word
		  if (! Substr (Command, "type")) 
			{
			  std::cout << "Illegal command.";
			  break;
			}
		  Cur_poly = Create_poly (Operand.c_str(), &A.Cur_poly_len);
		  if (Cur_poly == NULL) 
			{
			  std::cout << Operand << " is not a nonassociative word.";
			  break;
			}
		  else if (Cur_poly->terms->next != NULL)
			{
			  std::cout << Operand << " is not a nonassociative word.";
			}
		  else 
			{
			  i = Assoc_number(Cur_poly->terms->term);
			  std::cout << "The association type of the word = " << i << ".";
			}
		  break;
	  
		case '#':
		  // comment; ignore it (for batch processing)
		  break;

		case 'q':
		  // "quit". Quit Albert.
		  if (! Substr (Command, "quit")) 
			{
			  std::cout << "Illegal command.";
			  break;
			}
		  quit = 1;
		  break;
		  
		case '\0':
		  break;
		  
		default :
		  std::cout << "Illegal command." << std::endl;
		  break;
        }
    } // end of the main loop
  
 END:
  //free (Command);
  return (0);
} // end of main()




// Check if the polynomial has the compatible problem type.
// i.e if generators are 3a2b2c then each term in a polynomial should have at 
// most 3a's,2b's and 2c's. Else we can't use the multiplication table to answer
// the query whether the poly is an identity or not. 
static bool Compatible (const polynomial* Poly, const P_type ptype)
{
  if (Poly == NULL)
	{
	  return (0);
	}
  for (unsigned int i=0; i<NUM_LETTERS; i++)
	{
	  if (Poly->deg_letter[i] > ptype.degrees[i])
		{
		  return (0);
		}
	}
  return (1);
}


static void usage()
{
  std::cout << "Usage: albert [-a dirname]" << std::endl;
}


// wrapper for readline functions;
// returns line length, or -1 if error reading stdin occured
static size_t __getline (std::string& s)
{
  char* line = readline ("-->");
  if (line == NULL)
    {
      // end of stdin; this should happen in a non-interacive session only
      return (-1);
    }
  if (*line)
    {
      add_history (line);
    }
  size_t line_len = strlen (line);
  s = line;
  free (line);
  return (line_len);
}



// Used to check if Str1 is a string which is a prefix of Str2
static bool Substr (const std::string& Str1, const std::string& Str2)
{
  size_t len_str1 = Str1.length();
  size_t len_str2 = Str2.length();
  size_t i = 0;
  while ((i<len_str1) && (i<len_str2)) 
	{
	  if (Str1[i] != Str2[i])
		{
		  return (0);
		}
	  else
		{
		  i++;
		}
    }
  return (i >= len_str1);
}


// end of driver.cc
